@@ -26,6 +26,7 @@
 #include <config.h>
 #endif
 
+#include <libgtkhtml/gtkhtml.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <gtk/gtk.h>
 #include <bonobo/bonobo-main.h>
@@ -84,15 +85,20 @@ static GdkPixbuf * window_load_icon               (void);
 static void        window_set_sections            (YelpWindow        *window,
 						   GtkTreeModel      *sections);
 static void        window_handle_uri              (YelpWindow        *window,
-						   YelpURI           *uri);
+						   YelpURI           *uri,
+						   gboolean           historyq);
 static gboolean    window_handle_pager_uri        (YelpWindow        *window,
-						   YelpURI           *uri);
+						   YelpURI           *uri,
+						   gboolean           historyq);
 static gboolean    window_handle_html_uri         (YelpWindow        *window,
-						   YelpURI           *uri);
+						   YelpURI           *uri,
+						   gboolean           historyq);
 static void        window_handle_page             (YelpWindow        *window,
 						   YelpPage          *page,
 						   YelpURI           *uri);
 static void        window_disconnect              (YelpWindow        *window);
+static void        yelp_window_destroyed          (GtkWidget         *window,
+						   gpointer           user_data);
 
 static void        pager_contents_cb              (YelpPager         *pager,
 						   gpointer           user_data);
@@ -183,9 +189,6 @@ enum {
 static gint signals[LAST_SIGNAL] = { 0 };
 
 struct _YelpWindowPriv {
-    GNode          *doc_tree;
-    GList          *index;
-
     GtkWidget      *main_box;
     GtkWidget      *pane;
     GtkWidget      *side_sects;
@@ -225,6 +228,22 @@ struct _YelpWindowPriv {
     gchar          *toc;
 };
 
+typedef struct _IdleWriterContext IdleWriterContext;
+struct _IdleWriterContext {
+    YelpWindow   *window;
+
+    enum {
+	IDLE_WRITER_MEMORY,
+	IDLE_WRITER_VFS
+    } type;
+
+    const gchar  *buffer;
+    gint          cur;
+    gint          length;
+};
+
+static gboolean    idle_write    (IdleWriterContext *context);
+
 static GtkItemFactoryEntry menu_items[] = {
     {N_("/_File"),                   NULL, 0, 0, "<Branch>"     },
     {N_("/File/_New window"),        NULL,
@@ -358,13 +377,14 @@ yelp_window_new (GNode *doc_tree, GList *index)
 
     priv   = window->priv;
 
-    priv->doc_tree = doc_tree;
-    priv->index    = index;
-
     window_populate (window);
 
     gtk_window_set_icon (GTK_WINDOW (window), window_load_icon ());
 
+    g_signal_connect (G_OBJECT (window), "destroy",
+		      G_CALLBACK (yelp_window_destroyed),
+		      NULL);
+
     return GTK_WIDGET (window);
 }
 
@@ -388,7 +408,8 @@ yelp_window_open_uri (YelpWindow  *window,
     priv = window->priv;
 
     cur_uri = yelp_history_get_current (window->priv->history);
-    if (cur_uri && gnome_vfs_uri_equal (cur_uri->uri, uri->uri)) {
+    if ((cur_uri && cur_uri->uri && uri->uri)
+	&& gnome_vfs_uri_equal (cur_uri->uri, uri->uri)) {
 	const gchar *cur_frag, *frag;
 	cur_frag = gnome_vfs_uri_get_fragment_identifier (cur_uri->uri);
 	frag     = gnome_vfs_uri_get_fragment_identifier (uri->uri);
@@ -398,14 +419,13 @@ yelp_window_open_uri (YelpWindow  *window,
 	    return;
     }
 
-    yelp_history_goto (window->priv->history, uri);
-
-    window_handle_uri (window, uri);
+    window_handle_uri (window, uri, TRUE);
 }
 
 void
 window_handle_uri (YelpWindow  *window,
-		   YelpURI     *uri)
+		   YelpURI     *uri,
+		   gboolean     historyq)
 {
     YelpWindowPriv *priv;
     GError         *error = NULL;
@@ -422,19 +442,22 @@ window_handle_uri (YelpWindow  *window,
     case YELP_URI_TYPE_MAN:
     case YELP_URI_TYPE_INFO:
     case YELP_URI_TYPE_TOC:
-	handled = window_handle_pager_uri (window, uri);
+	handled = window_handle_pager_uri (window, uri, historyq);
 	break;
     case YELP_URI_TYPE_DOCBOOK_SGML:
 	yelp_set_error (&error, YELP_ERROR_NO_SGML);
 	break;
     case YELP_URI_TYPE_HTML:
-	handled = window_handle_html_uri (window, uri);
+	handled = window_handle_html_uri (window, uri, historyq);
 	break;
     case YELP_URI_TYPE_EXTERNAL:
 	str_uri = gnome_vfs_uri_to_string (uri->uri, GNOME_VFS_URI_HIDE_NONE);
 	gnome_url_show (str_uri, &error);
 	g_free (str_uri);
 	break;
+    case YELP_URI_TYPE_MAILTO:
+	gnome_url_show (uri->src_uri, &error);
+	break;
     case YELP_URI_TYPE_ERROR:
     default:
 	yelp_set_error (&error, YELP_ERROR_NO_DOC);
@@ -578,8 +601,6 @@ window_populate (YelpWindow *window)
     gtk_box_pack_start (GTK_BOX (priv->main_box),
 			priv->html_sw,
 			TRUE, TRUE, 0);
-
-    gtk_widget_grab_focus (yelp_html_get_widget (priv->html_view));
 }
 
 static GtkWidget *
@@ -711,7 +732,8 @@ window_set_sections (YelpWindow   *window,
 
 static gboolean
 window_handle_pager_uri (YelpWindow  *window,
-			 YelpURI     *uri)
+			 YelpURI     *uri,
+			 gboolean     historyq)
 {
     YelpWindowPriv *priv;
     GError         *error = NULL;
@@ -823,6 +845,9 @@ window_handle_pager_uri (YelpWindow  *window,
 			      window);
     }
 
+    if (historyq)
+	yelp_history_goto (window->priv->history, uri);
+
     window_set_sections (window,
 			 GTK_TREE_MODEL (yelp_pager_get_sections (pager)));
 
@@ -834,8 +859,6 @@ window_handle_pager_uri (YelpWindow  *window,
 	gchar *loading = _("Loading...");
 	GdkCursor *cursor = gdk_cursor_new (GDK_WATCH);
 
-	yelp_html_clear (priv->html_view);
-
 	gdk_window_set_cursor (GTK_WIDGET (window)->window, cursor);
 	gdk_cursor_unref (cursor);
 
@@ -858,6 +881,7 @@ window_handle_pager_uri (YelpWindow  *window,
 	gtk_window_set_title (GTK_WINDOW (window),
 			      (const gchar *) loading);
 
+	yelp_html_clear (priv->html_view);
 	yelp_html_printf
 	    (priv->html_view,
 	     "<html><head><meta http-equiv='Content-Type'"
@@ -898,7 +922,8 @@ window_handle_pager_uri (YelpWindow  *window,
 
 static gboolean
 window_handle_html_uri (YelpWindow    *window,
-			YelpURI       *uri)
+			YelpURI       *uri,
+			gboolean       historyq)
 {
     GnomeVFSHandle  *handle;
     GnomeVFSResult   result;
@@ -911,6 +936,9 @@ window_handle_html_uri (YelpWindow    *window,
     g_return_val_if_fail (YELP_IS_WINDOW (window), FALSE);
     g_return_val_if_fail (uri != NULL, FALSE);
 
+    if (historyq)
+	yelp_history_goto (window->priv->history, uri);
+
     window_set_sections (window, NULL);
 
     menu_item =
@@ -930,23 +958,28 @@ window_handle_html_uri (YelpWindow    *window,
 	gtk_widget_set_sensitive (menu_item, FALSE);
 
 
-    yelp_html_clear (priv->html_view);
-    yelp_html_set_base_uri (priv->html_view, uri);
-
     result = gnome_vfs_open (&handle,
 			     gnome_vfs_uri_get_path (uri->uri),
 			     GNOME_VFS_OPEN_READ);
 
     if (result != GNOME_VFS_OK) {
-	// FIXME: Give an error
+	GError *error = NULL;
+	yelp_set_error (&error, YELP_ERROR_NO_DOC);
+	window_error (window, error);
+	g_error_free (error);
 	return FALSE;
     }
 
+    yelp_html_clear (priv->html_view);
+    yelp_html_set_base_uri (priv->html_view, uri);
+
     while ((result = gnome_vfs_read
 	    (handle, buffer, BUFFER_SIZE, &n)) == GNOME_VFS_OK) {
 	yelp_html_write (priv->html_view, buffer, n);
     }
 
+    yelp_html_close (priv->html_view);
+
     gnome_vfs_close (handle);
 
     return TRUE;
@@ -955,7 +988,7 @@ window_handle_html_uri (YelpWindow    *window,
 static void
 window_handle_page (YelpWindow   *window,
 		    YelpPage     *page,
-		    YelpURI  *uri)
+		    YelpURI      *uri)
 {
     GtkWidget      *menu_item;
     GtkTreeModel   *model;
@@ -963,6 +996,7 @@ window_handle_page (YelpWindow   *window,
     YelpWindowPriv *priv;
     gchar          *id;
     gboolean        valid;
+    IdleWriterContext *context;
 
     g_return_if_fail (YELP_IS_WINDOW (window));
 
@@ -978,12 +1012,21 @@ window_handle_page (YelpWindow   *window,
 				-1);
 	    if (yelp_pager_uri_is_page (priv->pager, id, uri)) {
 		GtkTreePath *path = gtk_tree_model_get_path (model, &iter);
+		GtkTreeSelection *selection =
+		    gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->side_sects));
+
+		g_signal_handlers_block_by_func (selection,
+						 tree_selection_changed_cb,
+						 window);
 
 		gtk_tree_view_expand_to_path (GTK_TREE_VIEW (priv->side_sects),
 					      path);
 		gtk_tree_view_set_cursor (GTK_TREE_VIEW (priv->side_sects),
 					  path, NULL, FALSE);
 
+		g_signal_handlers_unblock_by_func (selection,
+						   tree_selection_changed_cb,
+						   window);
 		gtk_tree_path_free (path);
 		break;
 	    }
@@ -992,39 +1035,43 @@ window_handle_page (YelpWindow   *window,
 	}
     }
 
-    if (page->prev) {
-	priv->prev = page->prev;
-	menu_item =
-	    gtk_item_factory_get_item_by_action (priv->item_factory,
-						 YELP_WINDOW_GO_PREVIOUS);
-	if (menu_item)
-	    gtk_widget_set_sensitive (menu_item, TRUE);
-    }
-    if (page->next) {
-	priv->next = page->next;
-	menu_item =
-	    gtk_item_factory_get_item_by_action (priv->item_factory,
-						 YELP_WINDOW_GO_NEXT);
-	if (menu_item)
-	    gtk_widget_set_sensitive (menu_item, TRUE);
-    }
-    if (page->toc) {
-	priv->toc = page->toc;
-	menu_item =
-	    gtk_item_factory_get_item_by_action (priv->item_factory,
-						 YELP_WINDOW_GO_TOC);
-	if (menu_item)
-	    gtk_widget_set_sensitive (menu_item, TRUE);
-    }
+    priv->prev = page->prev;
+    menu_item =
+	gtk_item_factory_get_item_by_action (priv->item_factory,
+					     YELP_WINDOW_GO_PREVIOUS);
+    if (menu_item)
+	gtk_widget_set_sensitive (menu_item,
+				  priv->prev ? TRUE : FALSE);
+
+    priv->next = page->next;
+    menu_item =
+	gtk_item_factory_get_item_by_action (priv->item_factory,
+					     YELP_WINDOW_GO_NEXT);
+    if (menu_item)
+	gtk_widget_set_sensitive (menu_item,
+				  priv->next ? TRUE : FALSE);
+
+    priv->toc = page->toc;
+    menu_item =
+	gtk_item_factory_get_item_by_action (priv->item_factory,
+					     YELP_WINDOW_GO_TOC);
+    if (menu_item)
+	gtk_widget_set_sensitive (menu_item,
+				  priv->toc ? TRUE : FALSE);
 
     gtk_window_set_title (GTK_WINDOW (window),
 			  (const gchar *) page->title);
 
+    context = g_new0 (IdleWriterContext, 1);
+    context->window = window;
+    context->type   = IDLE_WRITER_MEMORY;
+    context->buffer = page->chunk;
+    context->length = strlen (page->chunk);
+
     yelp_html_clear (priv->html_view);
     yelp_html_set_base_uri (priv->html_view, uri);
-    yelp_html_write (priv->html_view,
-		     page->chunk,
-		     strlen (page->chunk));
+
+    gtk_idle_add ((GtkFunction) idle_write, context);
 
     if (gnome_vfs_uri_get_fragment_identifier (uri->uri)) {
 	yelp_html_jump_to_anchor
@@ -1063,6 +1110,22 @@ window_disconnect (YelpWindow *window)
     }
 }
 
+static void
+yelp_window_destroyed (GtkWidget *window,
+		       gpointer   user_data)
+{
+    YelpWindowPriv *priv;
+
+    g_return_if_fail (YELP_IS_WINDOW (window));
+
+    priv = YELP_WINDOW(window)->priv;
+
+    g_object_unref (priv->pane);
+    g_object_unref (priv->side_sw);
+    g_object_unref (priv->html_sw);
+    g_object_unref (priv->pager);
+}
+
 static void
 pager_contents_cb (YelpPager   *pager,
 		   gpointer     user_data)
@@ -1153,7 +1216,6 @@ html_uri_selected_cb (YelpHtml  *html,
 
     if (!handled) {
 	yelp_window_open_uri (window, uri);
-	yelp_uri_unref (uri);
     }
 }
 
@@ -1400,7 +1462,7 @@ window_history_action (YelpWindow        *window,
     }
 	
     if (uri) {
-	window_handle_uri (window, uri);
+	window_handle_uri (window, uri, FALSE);
     }
 }
 
@@ -1671,3 +1733,48 @@ tree_model_iter_following (GtkTreeModel  *model,
     g_assert_not_reached ();
     return FALSE;
 }
+
+/* Writing incrementally in an idle function has a number of advantages.  First,
+ * it keeps the interface responsive for really big pages.  Second, it prevents
+ * some weird rendering artifacts in gtkhtml2.  Third, and most important, it
+ * helps me beat the relayout race condition that I discuss at length in a big
+ * comment in yelp-html-gtkhtml2.c.
+ */
+
+static gboolean
+idle_write (IdleWriterContext *context)
+{
+    YelpWindowPriv *priv;
+
+    g_return_val_if_fail (context != NULL, FALSE);
+    g_return_val_if_fail (context->window != NULL, FALSE);
+
+    priv = context->window->priv;
+
+    switch (context->type) {
+    case IDLE_WRITER_MEMORY:
+	if (context->cur + BUFFER_SIZE < context->length) {
+	    yelp_html_write (priv->html_view,
+			     context->buffer + context->cur,
+			     BUFFER_SIZE);
+	    context->cur += BUFFER_SIZE;
+	    if (context->cur + BUFFER_SIZE >= context->length)
+		gtk_widget_grab_focus (yelp_html_get_widget (priv->html_view));
+	    return TRUE;
+	} else {
+	    if (context->length > context->cur)
+		yelp_html_write (priv->html_view,
+				 context->buffer + context->cur,
+				 context->length - context->cur);
+	    yelp_html_close (priv->html_view);
+	    g_free (context);
+	    return FALSE;
+	}
+	break;
+    case IDLE_WRITER_VFS:
+    default:
+	g_assert_not_reached ();
+    }
+
+    return FALSE;
+}
